#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains methods for generalized note alignment.
e.g.:
- matching a performance to a score without knowing the repeat structure 
- matching a monophonic voice to a full performance
"""

from .repeat_identification import RepeatIdentifier
from .subpart import SubPartMatcher
